-- Practical typed lazy contracts
-- Implementation without informative error messages and without Template Haskell
-- demonstrates how contracts work
-- Provides simple blaming of contract argument or caller as culprit. 

module Development.ContractSimpleBlame where

import Data.Maybe(maybe)
import Control.Monad(mplus)

data Partner = Client | Server | Contract deriving Show

type Blame = (Partner, Partner)
  -- first is culprit to blame, second is other partner

-- for contra-variant subcontract swap blame partners
contra :: Blame -> Blame
contra (p1,p2) = (p2,p1)

newClient :: Blame -> Partner -> Blame
newClient (p1,_) p = (p1,p)

type Contract a = a -> Blame -> Either Partner a

assert :: Contract a -> (a -> a)
assert = monitor (Server, Client)

monitor :: Blame -> Contract a -> (a -> a)
monitor b c x = case c x b of
  Right x  -> x
  Left p   -> error ("Contract failed. The " ++ show p ++ " is to blame.")

attach :: a -> Contract a -> a
attach x c = assert c x


(&) :: Contract a -> Contract a -> Contract a
c1 & c2 = \x b -> case c1 x b of
                    Right y -> c2 y b
                    Left p  -> Left p

(|>) :: Contract a -> Contract a -> Contract a
c1 |> c2 = \x b -> case c1 x b of
                     Right y -> Right y
                     Left _  -> c2 x b

(>->) :: Contract a -> Contract b -> Contract (a -> b)
pre >-> post = \f b -> Right (monitor b post . f . monitor (contra b) pre)

-- The dependent function contract combinator does not protect laziness!
(>>->) :: Contract a -> (a -> Contract b) -> Contract (a -> b)
pre >>-> post = 
  \f b -> Right (f `seq` \x -> let b' = contra b
                               in monitor b 
                                 (post (monitor (newClient b' Contract) pre x)) 
                                 (f (monitor b' pre x)))


true :: Contract a
true = \x b -> Right x

false :: Contract a
false = \x b -> cFail (contra b)


class Flat a where
  prop :: (a -> Bool) -> Contract a
  prop p = \x b -> if p x then Right x else cFail b

instance Flat Int
instance Flat Integer
instance Flat Float
instance Flat Double
instance Flat Char
instance Flat Bool

pNil :: Contract [a]
pNil [] b = Right []
pNil (_:_) b = cFail b

pCons :: Contract a -> Contract [a] -> Contract [a]
pCons cx cxs (z:zs) b = Right (monitor b cx z : monitor b cxs zs)
pCons cx cxs [] b = cFail b

pNotNil :: Contract [a]
pNotNil [] b = cFail b
pNotNil xs b = Right xs

pNotCons :: Contract [a]
pNotCons (_:_) b = cFail b
pNotCons xs b = Right xs

list :: Contract a -> Contract [a]
list c = pNil |> pCons c (list c)

io :: Contract a -> Contract (IO a)
io c m b = Right (m >>= return . monitor b c)

cFail :: Blame -> Either Partner a
cFail b = Left (fst b) 

p0 :: Contract a
p0 x b = Right x

p1 :: (a -> b) -> Contract a -> a -> Blame -> Either z b
p1 c c1 x1 b = Right (c (monitor b c1 x1))

p2 :: (a -> b -> c) -> Contract a -> Contract b -> 
      a -> b -> Blame -> Either z c
p2 c c1 c2 x1 x2 b = Right (c (monitor b c1 x1) (monitor b c2 x2))

p3 :: (a -> b -> c -> d) -> Contract a -> Contract b -> Contract c -> 
      a -> b -> c -> Blame -> Either z d
p3 c c1 c2 c3 x1 x2 x3 b = 
  Right (c (monitor b c1 x1) (monitor b c2 x2) (monitor b c3 x3))

pTuple0 :: Contract ()
pTuple0 () = p0  ()

pTuple2 :: Contract a -> Contract b -> Contract (a,b)
pTuple2 c1 c2 (x1,x2) = p2 (,) c1 c2 x1 x2

pTuple3 :: Contract a -> Contract b -> Contract c -> Contract (a,b,c)
pTuple3 c1 c2 c3 (x1,x2,x3) = p3 (,,) c1 c2 c3 x1 x2 x3

pNothing :: Contract (Maybe a)
pNothing Nothing = p0 Nothing
pNothing _ = cFail

pJust :: Contract a -> Contract (Maybe a)
pJust c (Just x) = p1 Just c x
pJust c _ = cFail

pNotNothing :: Contract (Maybe a)
pNotNothing Nothing = cFail
pNotNothing x = p0 x

pNotJust :: Contract (Maybe a)
pNotJust (Just _) = cFail
pNotJust x = p0 x

maybe :: Contract a -> Contract (Maybe a)
maybe c = pNothing |> pJust c 
